@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2023 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2024 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -417,13 +417,11 @@ static int oggdec(struct module_data *m, HIO_HANDLE *f, struct xmp_sample *xxs,
 	n = stb_vorbis_decode_memory(data, len, &ch, &rate, &pcm16);
 	free(data);
 
-	if (n <= 0) {
+	if (n <= 0 || ch != 1) {
 		free(pcm16);
 		return -1;
 	}
 
-	xxs->len = n;
-
 	if ((xxs->flg & XMP_SAMPLE_16BIT) == 0) {
 		uint8 *pcm = (uint8 *)pcm16;
 
@@ -437,6 +435,11 @@ static int oggdec(struct module_data *m, HIO_HANDLE *f, struct xmp_sample *xxs,
 		}
 		pcm16 = (int16 *)pcm;
 	}
+	if (xxs->flg & XMP_SAMPLE_STEREO) {
+		/* OXM stereo is a single channel non-interleaved stream. */
+		n >>= 1;
+	}
+	xxs->len = n;
 
 	flags |= SAMPLE_FLAG_NOLOAD;
 #ifdef WORDS_BIGENDIAN
@@ -684,7 +687,7 @@ static int load_instruments(struct module_data *m, int version, HIO_HANDLE *f)
 				xxs->lpe >>= 1;
 			}
 			if (xsh[j].type & XM_SAMPLE_STEREO) {
-				/* xxs->flg |= XMP_SAMPLE_STEREO; */
+				xxs->flg |= XMP_SAMPLE_STEREO;
 				xxs->len >>= 1;
 				xxs->lps >>= 1;
 				xxs->lpe >>= 1;
@@ -693,14 +696,15 @@ static int load_instruments(struct module_data *m, int version, HIO_HANDLE *f)
 			xxs->flg |= xsh[j].type & XM_LOOP_FORWARD ? XMP_SAMPLE_LOOP : 0;
 			xxs->flg |= xsh[j].type & XM_LOOP_PINGPONG ? XMP_SAMPLE_LOOP | XMP_SAMPLE_LOOP_BIDIR : 0;
 
-			D_(D_INFO "  size:%06x loop start:%06x loop end:%06x %c V%02x F%+04d P%02x R%+03d %s",
+			D_(D_INFO "  size:%06x loop start:%06x loop end:%06x %c V%02x F%+04d P%02x R%+03d %s%s",
 			   mod->xxs[sub->sid].len,
 			   mod->xxs[sub->sid].lps,
 			   mod->xxs[sub->sid].lpe,
 			   mod->xxs[sub->sid].flg & XMP_SAMPLE_LOOP_BIDIR ? 'B' :
 			   mod->xxs[sub->sid].flg & XMP_SAMPLE_LOOP ? 'L' : ' ',
 			   sub->vol, sub->fin, sub->pan, sub->xpo,
-			   mod->xxs[sub->sid].flg & XMP_SAMPLE_16BIT ? " (16 bit)" : "");
+			   mod->xxs[sub->sid].flg & XMP_SAMPLE_16BIT ? " (16 bit)" : "",
+			   xxs->flg & XMP_SAMPLE_STEREO ? " (stereo)" : "");
 		}
 
 		/* Read actual sample data */
@@ -742,12 +746,6 @@ static int load_instruments(struct module_data *m, int version, HIO_HANDLE *f)
 				} else {
 					total_sample_size += xsh[j].length;
 				}
-
-				/* TODO: implement stereo samples.
-				 * For now, just skip the right channel. */
-				if (xsh[j].type & XM_SAMPLE_STEREO) {
-					hio_seek(f, xsh[j].length >> 1, SEEK_CUR);
-				}
 			}
 		}
 
@@ -816,10 +814,6 @@ static int xm_load(struct module_data *m, HIO_HANDLE * f, const int start)
 		return -1;
 	}
 
-	if (xfh.restart > 255) {
-		D_(D_CRIT "bad restart position: %d", xfh.restart);
-		return -1;
-	}
 	if (xfh.channels > XMP_MAX_CHANNELS) {
 		D_(D_CRIT "bad channel count: %d", xfh.channels);
 		return -1;
@@ -852,7 +846,7 @@ static int xm_load(struct module_data *m, HIO_HANDLE * f, const int start)
 	mod->chn = xfh.channels;
 	mod->pat = xfh.patterns;
 	mod->ins = xfh.instruments;
-	mod->rst = xfh.restart;
+	mod->rst = xfh.restart >= xfh.songlen ? 0 : xfh.restart;
 	mod->spd = xfh.tempo;
 	mod->bpm = xfh.bpm;
 	mod->trk = mod->chn * mod->pat + 1;
@@ -1006,6 +1000,8 @@ static int xm_load(struct module_data *m, HIO_HANDLE * f, const int start)
 		libxmp_set_type(m, "ModPlug Tracker 1.16 XM %d.%02d",
 				xfh.version >> 8, xfh.version & 0xff);
 
+		m->quirk &= ~QUIRK_FT2BUGS;
+		m->flow_mode = FLOW_MODE_MPT_116;
 		m->mvolbase = 48;
 		m->mvol = 48;
 		libxmp_apply_mpt_preamp(m);
@@ -1016,7 +1012,7 @@ static int xm_load(struct module_data *m, HIO_HANDLE * f, const int start)
 		mod->xxc[i].pan = 0x80;
 	}
 
-	m->quirk |= QUIRKS_FT2;
+	m->quirk |= QUIRKS_FT2 | QUIRK_FT2ENV;
 	m->read_event_type = READ_EVENT_FT2;
 
 	return 0;
